// Copyright 2019 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import Foundation

/// This module synchronizes fuzzer instance in the same process.
///
/// TODO this module is currently mostly untested and should be regarded experimental.
/// TODO don't send back interesting programs to where they came from.

public class LocalWorker: Module {
    /// The master instance to synchronize with.
    private let master: Fuzzer
    
    public init(worker: Fuzzer, master: Fuzzer) {
        self.master = master
        
        master.async {
            // "Identify" with the master.
            master.dispatchEvent(master.events.WorkerConnected, data: worker.id)
            
            // Corpus synchronization
            master.registerEventListener(for: master.events.InterestingProgramFound) { ev in
                worker.async {
                    // Dropout can, if enabled in the fuzzer config, help workers become more independent
                    // from the rest of the fuzzers by forcing them to rediscover edges in different ways.
                    worker.importProgram(ev.program, enableDropout: true, origin: .master)
                }
            }
        }
        
        worker.registerEventListener(for: worker.events.CrashFound) { ev in
            master.async {
                master.importCrash(ev.program, origin: .worker(id: worker.id))
            }
        }
        
        worker.registerEventListener(for: worker.events.InterestingProgramFound) { ev in
            master.async {
                master.importProgram(ev.program, origin: .worker(id: worker.id))
            }
        }
        
        worker.registerEventListener(for: worker.events.Log) { ev in
            master.async {
                master.dispatchEvent(master.events.Log, data: ev)
            }
        }
        
        // Regularly send local statistics to the master
        if let stats = Statistics.instance(for: worker) {
            worker.timers.scheduleTask(every: 60 * Seconds) {
                let data = stats.compute()
                master.async {
                    Statistics.instance(for: master)?.importData(data, from: worker.id)
                }
            }
        }
    }
}
